import os
import traceback
import uuid
from typing import Type, Union

from django.core.files.storage import default_storage
from django.db import models
from django.db.models import FileField
from django.utils.deconstruct import deconstructible

ULPTR = Union['webproj.conventions.models.LanguagePhrase',
              'webproj.conventions.models.LanguageText',
              'webproj.conventions.models.LanguageRichText',
              ]


@deconstructible
class RandomFileName(object):
    def __init__(self, path):
        self.path = os.path.join(path, "%s%s")

    def __call__(self, _, filename):
        # @note It's up to the validators to check if it's the correct file type in name or if one even exist.
        extension = os.path.splitext(filename)[1]
        return self.path % (uuid.uuid4(), extension)


def file_cleanup(sender, instance, **_):
    for field in sender._meta.get_fields():
        if field and isinstance(field, FileField):
            default_storage.delete(getattr(instance, field.name).path)


class IntListField(models.CharField):
    description = "An integer list"

    def __init__(self, *args, **kwargs):
        if 'max_length' not in kwargs:
            kwargs['max_length'] = 255
        if 'default' not in kwargs:
            kwargs['default'] = ''
        super().__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        if value.strip() == '':
            return list()
        return list(map(int, map(str.strip, value.split(','))))

    def to_python(self, value):
        if isinstance(value, list):
            return value
        if value is None:
            return value
        for i in '[](){}':
            value = value.replace(i, '')
        if value.strip() == '':
            return list()
        return list(map(int, map(str.strip, value.split(','))))

    def get_prep_value(self, value):
        return ','.join(list(map(str, value)))


class FloatListField(models.TextField):
    description = "A float list"

    def __init__(self, *args, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = ''
        super().__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        if value.strip() == '':
            return list()
        return list(map(float, map(str.strip, value.split(','))))

    def to_python(self, value):
        if isinstance(value, list):
            return value
        if value is None:
            return value
        for i in '[](){}':
            value = value.replace(i, '')
        if value.strip() == '':
            return list()
        return list(map(float, map(str.strip, value.split(','))))

    def get_prep_value(self, value):
        return ','.join(list(map(str, value)))


def update_string_remidable_from_reminder(rmdb: ULPTR, text_en: str, text_es: str, text_pt: str) -> ULPTR:
    rmdb_needs_saving = False
    if (not hasattr(rmdb, 'en') or rmdb.en != text_en) and text_en is not None:
        rmdb.en = text_en
        rmdb_needs_saving = True
    if (not hasattr(rmdb, 'es') or rmdb.es != text_es) and text_es is not None:
        rmdb.es = text_es
        rmdb_needs_saving = True
    if (not hasattr(rmdb, 'pt') or rmdb.pt != text_pt) and text_pt is not None:
        rmdb.pt = text_pt
        rmdb_needs_saving = True
    # print(rmdb.__dict__)
    if rmdb_needs_saving:
        rmdb.save()
    return rmdb


def get_string_remidable_from_reminder_updating(model: Type[ULPTR], reminder: str, text_en: str, text_es: str, text_pt: str) -> ULPTR:
    (text_en, text_es, text_pt) = [x.strip(
        ' \r\n\t*') if x is not None else None for x in (text_en, text_es, text_pt)]
    if (rmdb := model.objects.filter(reminder_alias=reminder).first()) is not None:
        return update_string_remidable_from_reminder(rmdb, text_en, text_es, text_pt)
    else:
        rmdb = model(reminder_alias=reminder,
                     pt=text_pt, en=text_en, es=text_es)
        rmdb.save()
        return rmdb


def get_color_remidable_from_reminder_updating(model, reminder, hexcolor, opacity=255):
    if model.objects.filter(reminder_alias=reminder).exists():
        rmdb = model.objects.get(reminder_alias=reminder)
        rmdb_needs_saving = False
        if not hasattr(rmdb, 'color') or rmdb.color != hexcolor:
            rmdb.color = hexcolor
            rmdb_needs_saving = True
        if not hasattr(rmdb, 'opacity') or rmdb.opacity != opacity:
            rmdb.opacity = opacity
            rmdb_needs_saving = True
        if rmdb_needs_saving:
            rmdb.save()
        return rmdb
    else:
        rmdb = model()
        rmdb.reminder_alias = reminder
        rmdb.color = hexcolor
        rmdb.opacity = opacity
        rmdb.save()
        return rmdb
