from pathlib import Path

from django.conf import settings
from django.core.management.base import BaseCommand

from ....adminModelRegister import registrable_models_in_module
from ... import models


class Command(BaseCommand):
    help = 'Fixes ownership on the database'

    def add_arguments(self, _):
        return None

    def handle(self, *args, **options):
        self._fix_images()
        self._collect_garbage()
        self.stdout.write(self.style.SUCCESS('Script has finished'))

    def _fix_images(self):
        owned_images = list()
        for lii in models.LanguageInvariantImage.objects.all():
            lip = Path(lii.picture.path)
            if not lip.exists():
                self.stderr.write(self.style.ERROR(
                    f'Deleted image reference {lii.resource_uuid.hex} on database: {lip} does not exist on filesystem'))
                lii.delete()
            else:
                owned_images.append(lip)
        orphanated_messages = list()
        for i in Path('media').glob('*'):
            broken = False
            for j in owned_images:
                if i.samefile(j) or (i.suffix == '.gz' and Path(str(i)[:-3]).exists() and Path(str(i)[:-3]).samefile(j)):
                    broken = True
                    break
            if not broken:
                orphanated_messages.append(f'Orphan file on filesystem: {i}')
                i.unlink()
        del owned_images
        for orphanated_message in sorted(orphanated_messages):
            self.stderr.write(self.style.WARNING(orphanated_message))
        del orphanated_messages

    def _collect_garbage(self):
        mdls = registrable_models_in_module(models)
        mdlsd = {c.__name__: c for c in mdls}
        uncollectable_types = [
            models.User, *filter(lambda c: getattr(c, '_uncollectable', False), mdls)]
        strength_source_types = [
            *filter(lambda c: getattr(c, '_strength', False), mdls)]
        references_weakly = [
            *filter(lambda c: getattr(c, '_references_weakly', False), mdls)]
        references_weakly_str = [*map(lambda c: c.__name__, references_weakly)]
        relations = list()
        for mdl in mdls:
            for fld in mdl._meta.get_fields():
                if isinstance(fld, (
                    models.models.ForeignKey,
                    # models.models.ManyToManyField,
                    models.models.OneToOneField,
                    models.models.ForeignObject
                )):
                    reversible = False
                    if isinstance(fld, (models.models.ManyToManyField, models.models.OneToOneField)):
                        reversible = True
                    this = fld
                    remt = fld.remote_field
                    thism = this.model.__name__
                    thisn = this.name
                    remtm = remt.model.__name__
                    remtn = remt.name
                    if thisn.endswith('+') and len(thisn) > 1:
                        raise ValueError(
                            f'Name your ManyToManyField relation: {remtm}.{remtn}')
                    if remtn.endswith('+') and len(remtn) > 1:
                        raise ValueError(
                            f'Name your ManyToManyField relation: {thism}.{thisn}')
                    if thism not in references_weakly_str and remtm not in references_weakly_str:
                        relations.append((
                            (thism, thisn),
                            (remtm, remtn),
                            (fld.null, fld.blank)
                        ))
                    if settings.DEBUG:
                        print(
                            '!!!' if reversible else '',
                            (
                                (thism, thisn),
                                (remtm, remtn)
                            )
                        )
        # strength property: a model becomes strong by referencing a strong model with no nulls
        strong_dependencies_changed = True
        while strong_dependencies_changed:
            strong_dependencies_changed = False
            for k in relations:
                if k[0][0] not in mdlsd or k[1][0] not in mdlsd:
                    continue
                this = mdlsd[k[0][0]]
                that = mdlsd[k[1][0]]
                if not (k[2][0] and k[2][1]) and that in strength_source_types and this not in strength_source_types:
                    strength_source_types.append(this)
                    strong_dependencies_changed = True
        strength_source_types = list(set(strength_source_types))
        weak_source_types = list(set(mdls)-set(strength_source_types))
        weak_source_types_str = sorted(
            list(map(lambda a: a.__name__, weak_source_types)))
        del strong_dependencies_changed
        # strength property: a strong model won't be garbage-collected
        uncollectable_types = list(
            set(uncollectable_types) | set(strength_source_types))
        for i in weak_source_types:
            if not hasattr(i, 'copy'):
                raise ValueError(
                    f'Class {repr(i.__name__)} is weak and does not have copy() method')
        for wst in weak_source_types_str:
            wstm = mdlsd[wst]
            if wstm in uncollectable_types:
                continue
            for obj in wstm.objects.all():
                usages = list()
                for ((tm, tn), (rm, _), _) in relations:
                    if rm == wst:
                        tmm = mdlsd[tm]
                        for parent in tmm.objects.filter(**{tn: obj}).all():
                            usages.append((parent, tn, obj))
                if len(usages) == 1:
                    pass  # No problem here
                elif len(usages) < 1:
                    # No usages; orphan
                    self.stderr.write(self.style.ERROR(
                        f'Deleted {repr(wst)} object {obj.resource_uuid.hex} from database: orphan'))
                    self.stdout.write(self.style.ERROR(
                        f'Orphan data {type(obj).__name__}: {obj.__dict__}'))
                    obj.delete()
                else:
                    # Used more than once; needs review
                    self.stderr.write(self.style.ERROR(
                        f'Replicating {repr(wst)} object {obj.resource_uuid.hex} on database: {len(usages)} uses'))
                    self.stderr.write(self.style.WARNING(
                        f'  Doesn\'t touch {repr(type(usages[0][0]).__name__+"."+usages[0][1])} (at object {usages[0][0].resource_uuid.hex})'))
                    for usage in usages[1:]:
                        cp = usage[2].copy()
                        fd = usage[1]
                        pr = usage[0]
                        setattr(pr, fd, cp)
                        self.stderr.write(self.style.WARNING(
                            f'  Assign to {repr(type(pr).__name__+"."+fd)} (at object {pr.resource_uuid.hex}): {cp.resource_uuid.hex}'))
                        cp.save()
                        pr.save()
