# -*- encoding: utf-8 -*-

import uuid
from . import models
from typing import Optional


class DjangoModelConversorToNativeStructures:
    @staticmethod
    def language_string(language_string: models.LanguageString) -> dict:
        if language_string is None:
            return {'uuid': None}
        return {
            'alias': language_string.reminder_alias,
            'uuid': language_string.resource_uuid.hex,
            'pt': language_string.pt,
            'en': language_string.en,
            'es': language_string.es,
        }

    @staticmethod
    def language_image(language_image: models.LanguageImage) -> dict:
        if language_image is None:
            return {'uuid': None}
        return {
            'alias': language_image.reminder_alias,
            'uuid': language_image.resource_uuid.hex,
            'pt': language_image.pt.url,
            'en': language_image.en.url,
            'es': language_image.es.url,
        }

    @staticmethod
    def social_media(social_media: models.SocialMedia) -> dict:
        return {
            'alias': social_media.reminder_alias,
            'name': DMCTNS.language_string(social_media.name),
            'url': DMCTNS.language_string(social_media.url),
            'color': DMCTNS.color(social_media.color),
            'icon': social_media.icon,
            'ordering': social_media.ordering,
        }

    @staticmethod
    def decompose_date(date) -> dict:
        return {
            'formatted': str(date),
            'yr': int(date.strftime('%Y')),
            'mo': int(date.strftime('%m')),
            'dy': int(date.strftime('%d')),
            'mn': int(date.strftime('%M')),
            'hr': int(date.strftime('%H')),
            'sc': int(date.strftime('%S')),
            'wd': int(date.strftime('%w')),
        }

    @staticmethod
    def timezone(tz) -> dict:
        return {
            'name': tz.reminder_alias,
            'hr': tz.hour,
            'mn': tz.minute,
        }

    @staticmethod
    def registration_tier(registration_tier: models.ConventionRegistrationTier):
        return {
            'level': registration_tier.level,
            'is_lowest': registration_tier.is_lowest,
            'uuid': registration_tier.resource_uuid.hex,
            'tier': DMCTNS.language_string(registration_tier.tier.title),
            'benefits': [
                DMCTNS.language_string(item.description)
                for item in registration_tier.benefits.all()
            ],
        }

    @staticmethod
    def summary_link(summary_link: models.SummaryLink):
        return {
            'uuid': summary_link.resource_uuid.hex,
            'ordering': summary_link.ordering,
            'name': DMCTNS.language_string(summary_link.name),
            'description': DMCTNS.language_string(summary_link.description),
            'icon': summary_link.icon,
            'url': DMCTNS.language_string(summary_link.url),
        }

    @staticmethod
    def color_scheme(color_scheme: Optional[models.ConventionColorScheme]):
        if color_scheme is None:
            ruid = uuid.UUID('00000000-0000-4000-0000-000000000000')
            return DMCTNS.color_scheme(models.ConventionColorScheme(resource_uuid=ruid))
        return {
            'alias': color_scheme.reminder_alias,
            'uuid': color_scheme.resource_uuid.hex,
            'primary_darker': color_scheme.primary_darker,
            'primary_regular': color_scheme.primary_regular,
            'primary_lighter': color_scheme.primary_lighter,
            'accent_color': color_scheme.accent_color,
        }

    @staticmethod
    def color(color: models.Color) -> dict:
        if color is None:
            return {'uuid': None}
        return {
            'alias': color.reminder_alias,
            'uuid': color.resource_uuid.hex,
            'color': color.color,
            'opacity': color.opacity,
        }

    @staticmethod
    def place(place: models.Place) -> dict:
        return {
            'uuid': place.resource_uuid.hex,
            'label': DMCTNS.language_string(place.name),
            'color': DMCTNS.color(place.color),
            'ordering': place.ordering,
        }

    @staticmethod
    def activity_category(activity_category: models.ActivityCategory) -> dict:
        return {
            'uuid': activity_category.resource_uuid.hex,
            'label': DMCTNS.language_string(activity_category.name),
            'color': DMCTNS.color(activity_category.color),
            'ordering': activity_category.ordering,
        }

    @staticmethod
    def additional_rule(additional_rule: models.AdditionalRules) -> dict:
        return {
            'uuid': additional_rule.resource_uuid.hex,
            'name': DMCTNS.language_string(additional_rule.name),
            'rules': DMCTNS.language_string(additional_rule.rules),
            'ordering': additional_rule.ordering,
        }

    @staticmethod
    def banner(banner: models.BannerChange) -> dict:
        return {
            'uuid': banner.resource_uuid.hex,
            'alias': banner.reminder_alias,
            'show_up_after': DMCTNS.decompose_date(banner.show_up_after),
            'hide_after': DMCTNS.decompose_date(banner.hide_after),
            'banner': DMCTNS.language_image(banner.banner),
        }

    @staticmethod
    def event(event: models.ConventionActivity) -> dict:
        return {
            'uuid': event.resource_uuid.hex,
            'conbook_key': event.conbook_key,
            'conbook_pages': event.conbook_pages,
            'title': DMCTNS.language_string(event.title),
            'subtitle': DMCTNS.language_string(event.subtitle),
            'description': DMCTNS.language_string(event.description),
            'time_start': DMCTNS.decompose_date(event.time_start),
            'time_end': DMCTNS.decompose_date(event.time_end),
            'places': [
                item.resource_uuid.hex
                for item in event.places.all()
            ],
            'categories': [
                item.resource_uuid.hex
                for item in event.categories.all()
            ],
            'attendable_by': [
                item.resource_uuid.hex
                for item in event.attendable_by.all()
            ],
            'language': event.language,
            'hidden_from_time_table': event.hidden_from_time_table,
            'picture': DMCTNS.language_image(event.picture),
            'word2vec': [*map(float, event.word2vec)],
        }

    @staticmethod
    def registration_link(link: models.ConventionRegistrationLink) -> dict:
        return {
            'uuid': link.resource_uuid.hex,
            'alias': link.reminder_alias,
            'ordering': link.ordering,
            'name': DMCTNS.language_string(link.name),
            'url': DMCTNS.language_string(link.url),
            'color': DMCTNS.color(link.color),
            'appears': DMCTNS.decompose_date(link.appears),
            'vanishes': DMCTNS.decompose_date(link.vanishes),
        }

    @staticmethod
    def map_image(map_image: models.MapImage) -> dict:
        return {
            'uuid': map_image.resource_uuid.hex,
            'alias': map_image.reminder_alias,
            'name': DMCTNS.language_string(map_image.name),
            'image': DMCTNS.language_image(map_image.image),
            'url': None if map_image.url is None else DMCTNS.language_string(map_image.url),
        }

    @staticmethod
    def announcement(announcement: models.Announcement) -> dict:
        return {
            'uuid': announcement.resource_uuid.hex,
            'title': DMCTNS.language_string(announcement.title),
            'body': DMCTNS.language_string(announcement.body),
            'link': DMCTNS.language_string(announcement.link),
            'created': DMCTNS.decompose_date(announcement.created),
            'modified': DMCTNS.decompose_date(announcement.modified),
        }

    @staticmethod
    def urgent_message(urgent_message: models.UrgentMessage) -> dict:
        return None if urgent_message is None else {
            'uuid': urgent_message.resource_uuid.hex,
            'card': {
                'name': DMCTNS.language_string(urgent_message.card_name),
                'description': DMCTNS.language_string(urgent_message.card_dscr),
                'icon': urgent_message.card_icon,
            },
            'tab': {
                'name': DMCTNS.language_string(urgent_message.tab_name),
                'description': DMCTNS.language_string(urgent_message.tab_dscr),
                'text': DMCTNS.language_string(urgent_message.tab_text),
            }
        }

    @staticmethod
    def convention_edition(convention_edition: models.ConventionEdition) -> dict:
        return {
            'uuid': convention_edition.resource_uuid.hex,
            'name': DMCTNS.language_string(convention_edition.name),
            'places': [
                DMCTNS.place(item)
                for item in models.Place.objects.filter(on_activities__in=convention_edition.events.all()).distinct()
            ],
            'tags': [
                DMCTNS.activity_category(item)
                for item in models.ActivityCategory.objects.filter(on_activities__in=convention_edition.events.all()).distinct()
            ],
            'lowest_registration_tier': (
                convention_edition.registration_tiers.order_by(
                    'level', '-is_lowest').all()[0].resource_uuid.hex
            ) if convention_edition.registration_tiers.exists() else None,
            'registration_tiers': [
                DMCTNS.registration_tier(item)
                for item in convention_edition.registration_tiers.order_by('level', '-is_lowest').all()
            ],
            'do_not_ask_registration_tier': convention_edition.do_not_ask_registration_tier,
            'theme': DMCTNS.language_string(convention_edition.theme),
            'hashtag_reminder': convention_edition.hashtag_reminder,
            'timezone': DMCTNS.timezone(convention_edition.timezone),
            'rule_621_checker': convention_edition.rule_621_checker,
            'next_edition': DMCTNS.decompose_date(convention_edition.next_edition_date),
            'ceremony_opening_time': DMCTNS.decompose_date(convention_edition.ceremony_opening_time),
            'ceremony_closing_time': DMCTNS.decompose_date(convention_edition.ceremony_closing_time),
            'split_day_into_n_parts': convention_edition.split_day_into_n_parts,
            'image_default_event': DMCTNS.language_image(convention_edition.image_default_event),
            'image_favorites': DMCTNS.language_image(convention_edition.image_favorites),
            'maps': [
                DMCTNS.map_image(item)
                for item in convention_edition.maps.all()
                if not item.hidden
            ],
            'language': convention_edition.language,
            'events': [
                DMCTNS.event(item)
                for item in convention_edition.events.filter(is_draft=False).all()
            ],
            'additional_rules': [
                DMCTNS.additional_rule(item)
                for item in convention_edition.additional_rules.all()
                if not item.hidden
            ],
            'registration_links': [
                DMCTNS.registration_link(item)
                for item in convention_edition.registration_links.all()
            ],
            'fire_notifications_n_minutes_before': 15,
            'convention_day_start_time':  DMCTNS.decompose_date(convention_edition.convention_day_start_time),
            'convention_day_end_time': DMCTNS.decompose_date(convention_edition.convention_day_end_time),
            'color_scheme': DMCTNS.color_scheme(getattr(convention_edition, 'color_scheme', None)),
            'summary_links': [
                DMCTNS.summary_link(item)
                for item in convention_edition.summary_links.all()
                if not item.hidden
            ],
        }

    @staticmethod
    def convention_series(convention_series: models.ConventionSeries) -> dict:
        announcements = [
            DMCTNS.announcement(announcement)
            for announcement in convention_series.announcements.all()
        ]
        if convention_series.announcement_display_limit > 0:
            announcements = announcements[:convention_series.announcement_display_limit]
        return {
            'uuid': convention_series.resource_uuid.hex,
            'name': DMCTNS.language_string(convention_series.name),
            'featured': None if convention_series.featured is None else convention_series.featured.resource_uuid.hex,
            'editions': ([
                DMCTNS.convention_edition(item)
                for item in convention_series.editions.all()
            ] if not convention_series.limit_availability_to_featured else [
                DMCTNS.convention_edition(
                    convention_series.featured or
                    convention_series.editions.all().first())
            ]),
            'timezone': DMCTNS.timezone(convention_series.timezone),
            'social_medias': [
                DMCTNS.social_media(item)
                for item in convention_series.social_medias.all()
            ],
            'default_banner': DMCTNS.language_image(convention_series.default_banner),
            'banners': [
                DMCTNS.banner(item)
                for item in convention_series.banners.all()
            ],
            'language': convention_series.language,
            'languages': convention_series.languages.split(','),
            'statute': DMCTNS.language_string(convention_series.statute),
            'startup_terms': DMCTNS.language_string(convention_series.startup_terms),
            'announcements': announcements,
            'urgent_message': DMCTNS.urgent_message(convention_series.urgent_message),
            'countdown_on_splash': convention_series.countdown_on_splash,
        }


DMCTNS = DjangoModelConversorToNativeStructures
