import traceback
import uuid
from io import StringIO
from pathlib import Path

from ckeditor.fields import RichTextField
from colorfield.fields import ColorField
from django import forms
from django.conf import settings
from django.contrib.admin.widgets import (AdminTextareaWidget,
                                          AdminTextInputWidget)
from django.contrib.auth import get_user_model
from django.core.validators import (MaxValueValidator, MinValueValidator,
                                    RegexValidator)
from django.db import models
from django.dispatch import receiver

from .fa_icons import icons_fa
from .helpers import IntListField, FloatListField, RandomFileName, file_cleanup
from .md_icons import icons_md

User = get_user_model()

HELP_FILL_TZ = "fill with the time on the timezone of the event."

TRILINGUAL_CHOICES = (
    ('pt', 'Portuguese'),
    ('en', 'English'),
    ('es', 'Spanish')
)

ALL_ICONS = [
    (None, '-- no icon --'),
    ('ic_telegram', 'Telegram'),
    ('ic_twitter', 'Twitter'),
    ('ic_facebook_official', 'Facebook'),
    ('ic_youtube_play', 'YouTube'),
    ('ic_instagram', 'Instagram'),
    ('ic_discord_logo_black', 'Discord'),
    ('ic_gamepad_24dp', 'Gamepad'),
    ('ic_baseline_videogame_asset_24', 'Videogame'),
    ('ic_web_black_24dp', 'Web browser'),
    *icons_fa,
    *icons_md,
]


class SingleLineTextField(models.TextField):
    def formfield(self, **kwargs):
        defaults = {'max_length': 999999999, 'form_class': forms.CharField}
        if self.null and not self.connection.features.interprets_empty_strings_as_nulls:
            defaults['empty_value'] = None
        defaults.update(kwargs)
        if defaults['widget'] is AdminTextareaWidget:
            defaults['widget'] = AdminTextInputWidget
        elif defaults['widget'] is forms.Textarea:
            defaults['widget'] = forms.TextInput
        return models.Field.formfield(self, **defaults)

# Create your models here.


class ConventionSeriesMixin:
    pass


class ConventionEditionMixin:
    pass


class CopyableMixin:
    def copy(self):
        return None


class Timestampable(models.Model):
    owner = models.ForeignKey(
        User, default=None, on_delete=models.SET_NULL, related_name='+', blank=True, null=True)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    resource_uuid = models.UUIDField(
        default=uuid.uuid4, editable=False, primary_key=False)

    class Meta:
        abstract = True

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'{type(self).__name__}(**{repr(self.__dict__)})'

    def on_post_save(self, sender, **kwargs):
        self._delete_json_cache()

    def on_post_delete(self, sender, **kwargs):
        self._delete_json_cache()

    def _delete_json_cache(self):
        jc = Path(settings.STATIC_ROOT).joinpath('json_cache')
        if jc.exists():
            for f in jc.glob('*'):
                if f.is_file():
                    f.unlink()

    def save(self, *args, **kwargs):
        if settings.DEBUG:
            sio = StringIO()
            traceback.print_stack(file=sio)
            print(f'saved {self}\n{sio.getvalue()}')
        return super().save(*args, **kwargs)


@receiver(models.signals.post_save)
def timestampable_on_post_save(sender, instance, **kwargs):
    if issubclass(sender, Timestampable):
        instance.on_post_save(sender, **kwargs)


@receiver(models.signals.post_delete)
def timestampable_on_post_delete(sender, instance, **kwargs):
    if issubclass(sender, Timestampable):
        instance.on_post_delete(sender, **kwargs)


class IRemindable(Timestampable):
    @property
    def reminder_title(self):
        return self.__class__.__name__

    def __str__(self):
        return f'{self.reminder_title} {self.kind} #{self.pk}: {self.reminder_alias}'

    class Meta:
        abstract = True


class Remindable(IRemindable):
    reminder_alias = models.CharField(default='', unique=False, max_length=255, validators=[
        RegexValidator(r'^[A-Za-z_][A-Za-z0-9_]*$'),
    ], help_text='All characters must be alphanumerical or underline, except by the first character, which must not be alphanumerical.')

    class Meta:
        abstract = True


class LanguageString(Remindable):
    resource_uuid = models.UUIDField(
        default=uuid.uuid4, editable=False, unique=True, primary_key=False)
    reminder_title = 'Language'
    untranslatable_pt = models.BooleanField(
        default=False, null=False, blank=True)
    untranslatable_en = models.BooleanField(
        default=False, null=False, blank=True)
    untranslatable_es = models.BooleanField(
        default=False, null=False, blank=True)

    class Meta:
        abstract = True
        ordering = ('reminder_alias', )


class SingleChildLanguagePhraseCopyable(LanguageString, CopyableMixin):
    _one_per_parent = True

    def copy(self):
        cp = type(self)(pt=self.pt, en=self.en, es=self.es)
        cp.reminder_alias = str(cp.resource_uuid.hex)
        return cp

    class Meta:
        abstract = True


class LanguagePhrase(SingleChildLanguagePhraseCopyable):
    kind = 'phrase'
    pt = SingleLineTextField(default='')
    en = SingleLineTextField(default='')
    es = SingleLineTextField(default='')


class LanguageText(SingleChildLanguagePhraseCopyable):
    kind = 'text'
    pt = models.TextField(default='')
    en = models.TextField(default='')
    es = models.TextField(default='')


class LanguageRichText(SingleChildLanguagePhraseCopyable):
    kind = 'rich text'
    pt = RichTextField(default='')
    en = RichTextField(default='')
    es = RichTextField(default='')


class SyncUrlForLanguageThing(IRemindable, CopyableMixin):
    kind = 'url syncer for ???'
    url = models.URLField(default='', null=False, blank=False, max_length=2048)
    clone_paused_until_gets_200_url = models.URLField(
        default=None, blank=True, null=True, max_length=2048)
    clone_paused_until_gets_200 = models.BooleanField(
        default=False, null=False, blank=True)
    # Don't forget to implement: '''associated_language = models.ForeignKey(to=SingleChildLanguagePhraseCopyable, null=False, blank=False)'''

    @property
    def reminder_alias(self):
        return self.associated_language.reminder_alias

    class Meta:
        abstract = True


class SyncUrlForLanguagePhrase(SyncUrlForLanguageThing):
    _uncollectable = True
    _references_weakly = True
    kind = 'url syncer for phrase'
    associated_language = models.ForeignKey(
        to=LanguagePhrase, null=False, blank=False, on_delete=models.CASCADE)


class SyncUrlForLanguageText(SyncUrlForLanguageThing):
    _uncollectable = True
    _references_weakly = True
    kind = 'url syncer for text'
    associated_language = models.ForeignKey(
        to=LanguageText, null=False, blank=False, on_delete=models.CASCADE)


class SyncUrlForLanguageRichText(SyncUrlForLanguageThing):
    _uncollectable = True
    _references_weakly = True
    kind = 'url syncer for rich text'
    associated_language = models.ForeignKey(
        to=LanguageRichText, null=False, blank=False, on_delete=models.CASCADE)


class LanguageInvariantImage(LanguageString):
    _strength = True
    kind = 'invariant image'
    picture = models.ImageField(upload_to=RandomFileName(''))
    convention = models.ForeignKey('ConventionSeries', default=None, null=True,
                                   blank=False, on_delete=models.CASCADE, related_name='raw_images')

    @property
    def pt(self): return self.picture

    @property
    def en(self): return self.picture

    @property
    def es(self): return self.picture

    @property
    def untranslatable_pt(self): return False

    @property
    def untranslatable_en(self): return False

    @property
    def untranslatable_es(self): return False

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


models.signals.post_delete.connect(file_cleanup, sender=LanguageInvariantImage,
                                   dispatch_uid="webproj.conventions.LanguageInvariantImage.file_cleanup")


class LanguageImage(SingleChildLanguagePhraseCopyable):
    kind = 'image'
    pt_image = models.ForeignKey(
        LanguageInvariantImage, on_delete=models.CASCADE, related_name='+')
    en_image = models.ForeignKey(
        LanguageInvariantImage, on_delete=models.CASCADE, related_name='+')
    es_image = models.ForeignKey(
        LanguageInvariantImage, on_delete=models.CASCADE, related_name='+')
    convention = models.ForeignKey('ConventionSeries', default=None, null=True,
                                   blank=False, on_delete=models.CASCADE, related_name='images')

    @property
    def pt(self): return self.pt_image.picture

    @property
    def en(self): return self.en_image.picture

    @property
    def es(self): return self.es_image.picture


class Announcement(Timestampable):
    convention = models.ForeignKey(
        'ConventionSeries', on_delete=models.CASCADE, related_name='announcements')
    title = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    body = models.ForeignKey(
        LanguageText, on_delete=models.CASCADE, related_name='+')
    link = models.ForeignKey(LanguagePhrase, blank=True,
                             null=True,  on_delete=models.CASCADE, related_name='+')
    telegram_id = models.IntegerField(default=-1, blank=True, null=False)

    def __str__(self):
        return f'Announcement #{self.pk}: {self.title}'

    class Meta:
        ordering = ('-created', '-telegram_id', '-pk')

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


class Color(Remindable, CopyableMixin):
    _one_per_parent = True
    color = ColorField(default='#000000')
    opacity = models.IntegerField(default=255, validators=[MinValueValidator(
        0), MaxValueValidator(255)], help_text="Opacity: 0 is transparent, 255 is opaque.")
    kind = ''

    def copy(self):
        cp = type(self)(color=self.color, opacity=self.opacity)
        cp.reminder_alias = str(cp.resource_uuid.hex)
        return cp


class ActivityCategory(Timestampable):
    edition = models.ForeignKey('ConventionEdition', blank=False,
                                null=True, on_delete=models.CASCADE, related_name='categories')
    name = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    color = models.ForeignKey(
        Color, on_delete=models.CASCADE, related_name='+')
    ordering = models.IntegerField(default=0, blank=True, null=False)

    def __str__(self):
        return f'Category #{self.pk}: {self.name.en}'

    class Meta:
        ordering = ('ordering', '-pk', )

    def _delete_json_cache(self):
        if self.edition is None:
            super()._delete_json_cache()
        else:
            self.edition._delete_json_cache()


class Place(Timestampable):
    edition = models.ForeignKey('ConventionEdition', blank=False,
                                null=True, on_delete=models.CASCADE, related_name='places')
    name = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    color = models.ForeignKey(
        Color, on_delete=models.CASCADE, related_name='+')
    ordering = models.IntegerField(default=0, blank=True, null=False)

    def __str__(self):
        return f'Place #{self.pk}: {self.name.en}'

    class Meta:
        ordering = ('ordering', '-pk', )

    def _delete_json_cache(self):
        if self.edition is None:
            super()._delete_json_cache()
        else:
            self.edition._delete_json_cache()


class Timezone(Remindable, CopyableMixin):
    _one_per_parent = True
    kind = ''
    hour = models.IntegerField(
        default=-3, validators=[MinValueValidator(-24), MaxValueValidator(24)])
    minute = models.IntegerField(default=0, validators=[
                                 MinValueValidator(0), MaxValueValidator(60)])

    def copy(self):
        cp = type(self)(hour=self.hour, minute=self.minute)
        cp.reminder_alias = str(cp.resource_uuid.hex)
        return cp


class RegistrationBenefitDefinition(Timestampable):
    edition = models.ForeignKey('ConventionEdition', blank=False, null=True,
                                on_delete=models.CASCADE, related_name='benefit_definitions')
    description = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    ordering = models.IntegerField(default=0, blank=True, null=False)

    def __str__(self):
        return f'Registration benefit definition #{self.pk}: {self.description.en}'

    class Meta:
        ordering = ('ordering', '-pk', )

    def _delete_json_cache(self):
        if self.edition is None:
            super()._delete_json_cache()
        else:
            self.edition._delete_json_cache()


class RegistrationTierDefinition(Timestampable):
    edition = models.ForeignKey('ConventionEdition', blank=False, null=True,
                                on_delete=models.CASCADE, related_name='tier_definitions')
    title = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    ordering = models.IntegerField(default=0, blank=True, null=False)

    def __str__(self):
        return f'Registration tier definition #{self.pk}: {self.title.en}'

    class Meta:
        ordering = ('ordering', '-pk', )

    def _delete_json_cache(self):
        if self.edition is None:
            super()._delete_json_cache()
        else:
            self.edition._delete_json_cache()


class ConventionRegistrationTier(Timestampable):
    level = models.IntegerField(default=0)
    tier = models.ForeignKey(RegistrationTierDefinition,
                             on_delete=models.CASCADE, related_name='+')
    benefits = models.ManyToManyField(
        RegistrationBenefitDefinition, blank=True, related_name='enables')
    convention = models.ForeignKey(
        'ConventionEdition', on_delete=models.CASCADE, related_name='registration_tiers')
    is_lowest = models.BooleanField(default=False, blank=True, null=False)

    def __str__(self):
        return f'Convention registration tier #{self.pk}: {self.convention.name.en} - {self.tier.title.en} ({self.benefits.count()})'

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


class ConventionActivity(Timestampable):
    is_draft = models.BooleanField(default=False, null=False, blank=True)
    conbook_key = models.CharField(default='', max_length=255)
    conbook_pages = IntListField()
    title = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    subtitle = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, blank=True, null=True, related_name='+')
    description = models.ForeignKey(
        LanguageText, on_delete=models.CASCADE, related_name='+')
    time_start = models.DateTimeField(help_text=HELP_FILL_TZ)
    time_end = models.DateTimeField(help_text=HELP_FILL_TZ)
    places = models.ManyToManyField(Place, related_name='on_activities')
    categories = models.ManyToManyField(
        ActivityCategory, blank=True, related_name='on_activities')
    attendable_by = models.ManyToManyField(
        ConventionRegistrationTier, related_name='allows_entrance_in')
    language = models.CharField(
        max_length=2,
        choices=TRILINGUAL_CHOICES,
        default='pt',
    )
    hidden_from_time_table = models.BooleanField(
        default=False, blank=True, null=False)
    picture = models.ForeignKey(LanguageInvariantImage, blank=True,
                                null=True, related_name='+', on_delete=models.SET_NULL)
    convention = models.ForeignKey(
        'ConventionEdition', related_name='events', on_delete=models.CASCADE)
    word2vec_original_data_hash = models.BinaryField(
        default=b'', max_length=64, null=False, blank=True)
    word2vec = FloatListField(default='', null=False, blank=True)
    remote_uuid = models.CharField(
        default=None, blank=True, null=True, max_length=255)

    class Meta:
        ordering = ('-time_start', '-time_end', 'conbook_key')
        indexes = [models.Index(fields=['convention']),
                   models.Index(fields=['remote_uuid']),
                   models.Index(fields=['convention', 'remote_uuid'])]

    def __str__(self):
        return f'Convention activity #{self.pk}: {self.convention.name.en} - {self.title.en} [{self.conbook_key}] ({self.time_start} - {self.time_end})'

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


class ConventionSeries(Timestampable):
    _uncollectable = True
    _strength = True
    name = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    url = models.SlugField(default='', unique=True, max_length=255)
    clone_type = models.CharField(
        default=None, blank=True, null=True, max_length=255)
    clone_from = models.CharField(
        default=None, blank=True, null=True, max_length=2048)
    clone_paused_until_gets_200_url = models.URLField(
        default=None, blank=True, null=True, max_length=2048)
    clone_paused_until_gets_200 = models.BooleanField(
        default=False, null=False, blank=True)
    languages = models.TextField(default=','.join(
        [l[0] for l in TRILINGUAL_CHOICES]), blank=False, null=False)
    featured = models.ForeignKey(
        'ConventionEdition', null=True, blank=True, on_delete=models.CASCADE, related_name='+')
    limit_availability_to_featured = models.BooleanField(
        default=False, blank=True, null=False)
    statute = models.ForeignKey(
        LanguageRichText, on_delete=models.CASCADE, related_name='+')
    startup_terms = models.ForeignKey(
        LanguageRichText, on_delete=models.SET_NULL, blank=True, null=True, related_name='+')
    timezone = models.ForeignKey(
        Timezone, on_delete=models.CASCADE, related_name='+')
    default_banner = models.ForeignKey(
        LanguageImage, default=None, blank=True, null=True, on_delete=models.CASCADE, related_name='+')
    telegram_channel = models.CharField(
        default="", blank=True, null=False, max_length=255)
    announcement_display_limit = models.IntegerField(
        default=0, blank=True, null=False)
    app_id_ios = models.CharField(
        default='', blank=True, null=False, max_length=255)
    app_id_android = models.CharField(
        default='', blank=True, null=False, max_length=255)
    app_url_pwa = models.URLField(default=None, blank=True, null=True)
    urgent_message = models.OneToOneField(
        'UrgentMessage', default=None, blank=True, null=True, on_delete=models.SET_NULL, related_name='parent')
    countdown_on_splash = models.BooleanField(
        default=True, blank=True, null=False)
    language = models.CharField(
        max_length=2,
        choices=TRILINGUAL_CHOICES,
        default='pt',
    )
    # banners ← foreign_relation
    # social_medias ← foreign_relation

    def __str__(self):
        return f'Convention series #{self.pk}: {self.name.en} (/{self.url})'

    def _delete_json_cache(self):
        jc = Path(settings.STATIC_ROOT).joinpath('json_cache')
        fl = jc.joinpath(f'convention_{self.url}.json')
        flz = jc.joinpath(f'convention_{self.url}.json.gz')
        if fl.exists():
            fl.unlink()
        if flz.exists():
            flz.unlink()


class UrgentMessage(Timestampable, CopyableMixin):
    card_name = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    card_dscr = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    card_icon = models.CharField(
        default='ic_warning', blank=False, null=False, max_length=255, choices=ALL_ICONS)
    tab_name = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    tab_dscr = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    tab_text = models.ForeignKey(
        LanguageRichText, on_delete=models.CASCADE, related_name='+')

    def __str__(self):
        if getattr(self, 'parent', None) is not None:
            return f'Urgent message #{self.pk} for: {self.parent.name.en} (/{self.parent.url})'
        else:
            return f'Urgent message #{self.pk}, at a orphan state'

    def _delete_json_cache(self):
        if getattr(self, 'parent', None) is None:
            super()._delete_json_cache()
        else:
            self.parent._delete_json_cache()

    def copy(self):
        return type(self)(
            owner=self.owner,
            card_name=self.card_name.copy(),
            card_dscr=self.card_dscr.copy(),
            card_icon=self.card_icon,
            tab_name=self.tab_name.copy(),
            tab_dscr=self.tab_dscr.copy(),
            tab_text=self.tab_text.copy()
        )


class SocialMedia(Remindable):
    ordering = models.IntegerField(default=0)
    convention = models.ForeignKey(
        ConventionSeries, on_delete=models.CASCADE, related_name='social_medias')
    name = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    url = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    icon = models.CharField(default='ic_web_black_24dp', blank=False,
                            null=False, max_length=255, choices=ALL_ICONS)
    kind = ''
    color = models.ForeignKey(
        Color, on_delete=models.CASCADE, related_name='+')

    class Meta:
        ordering = ('ordering', 'name__reminder_alias', '-pk', )

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


class ConventionEdition(Timestampable):
    # id ← inherited
    # events ← foreign_relation
    # registration_tiers ← foreign_relation
    # lowestRegistrationTier ← process
    # places ← process
    # tags ← process
    edition_of = models.ForeignKey(
        ConventionSeries, on_delete=models.CASCADE, related_name='editions')
    name = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    theme = models.ForeignKey(
        LanguagePhrase, on_delete=models.CASCADE, related_name='+')
    hashtag_reminder = models.CharField(default='', max_length=255)
    timezone = models.ForeignKey(
        Timezone, on_delete=models.CASCADE, related_name='+')
    convention_day_start_time = models.TimeField(
        help_text=HELP_FILL_TZ)
    convention_day_end_time = models.TimeField(
        help_text=HELP_FILL_TZ)
    rule_621_checker = models.BooleanField(
        default=False, blank=True, null=False)
    do_not_ask_registration_tier = models.BooleanField(
        default=False, blank=True, null=False)
    next_edition_date = models.DateField(
        help_text=HELP_FILL_TZ)
    ceremony_opening_time = models.DateTimeField(
        help_text=HELP_FILL_TZ)
    ceremony_closing_time = models.DateTimeField(
        help_text=HELP_FILL_TZ)
    image_default_event = models.ForeignKey(
        LanguageImage, default=None, null=True, blank=True, on_delete=models.CASCADE, related_name='+')
    image_favorites = models.ForeignKey(
        LanguageImage, default=None, null=True, blank=True, on_delete=models.CASCADE, related_name='+')
    split_day_into_n_parts = models.IntegerField(default=24, blank=False, null=False, validators=[
                                                 MaxValueValidator(1440), MinValueValidator(1)])
    language = models.CharField(
        max_length=2,
        choices=TRILINGUAL_CHOICES,
        default='pt',
    )

    def __str__(self):
        return f'Convention #{self.pk}: {self.name.en}'

    class Meta:
        ordering = ('ceremony_opening_time', '-pk', )

    def _delete_json_cache(self):
        if self.edition_of is None:
            super()._delete_json_cache()
        else:
            self.edition_of._delete_json_cache()


class ConventionColorScheme(Remindable):
    kind = 'color scheme'

    convention = models.OneToOneField(ConventionEdition, default=None, blank=False,
                                      null=False, on_delete=models.CASCADE, related_name="color_scheme")

    primary_darker = ColorField(default="#7f0000")
    primary_regular = ColorField(default="#b71c1c")
    primary_lighter = ColorField(default="#f05545")
    accent_color = ColorField(default="#f9a825")

    class Meta:
        ordering = ('-convention__pk', '-pk', )

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


class BannerChange(Remindable):
    convention = models.ForeignKey(ConventionSeries, default=None, blank=False,
                                   null=False, on_delete=models.CASCADE, related_name="banners")
    show_up_after = models.DateTimeField(
        help_text=HELP_FILL_TZ)
    hide_after = models.DateTimeField(
        help_text=HELP_FILL_TZ)
    banner = models.ForeignKey(
        LanguageImage, default=None, on_delete=models.CASCADE, related_name='+')

    reminder_title = 'Banner change'
    kind = ''

    class Meta:
        ordering = ('show_up_after', '-pk', )

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


class MapImage(Remindable):
    ordering = models.IntegerField(default=0)
    convention = models.ForeignKey(ConventionEdition, default=None, blank=False,
                                   null=False, on_delete=models.CASCADE, related_name="maps")
    name = models.ForeignKey(LanguagePhrase, blank=False,
                             null=False, on_delete=models.CASCADE, related_name='+')
    image = models.ForeignKey(LanguageImage, blank=False,
                              null=False, on_delete=models.CASCADE, related_name='+')
    url = models.ForeignKey(LanguagePhrase, blank=True,
                            null=True, on_delete=models.CASCADE, related_name='+')
    hidden = models.BooleanField(default=False, blank=True, null=False)
    reminder_title = 'Map'
    kind = 'image'

    class Meta:
        ordering = ('convention__ceremony_opening_time', 'ordering', '-pk', )

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


class ConventionRegistrationLink(Remindable):
    ordering = models.IntegerField(default=0)
    convention = models.ForeignKey(ConventionEdition, default=None, blank=False,
                                   null=False, on_delete=models.CASCADE, related_name="registration_links")
    name = models.ForeignKey(LanguagePhrase, blank=False,
                             null=False, on_delete=models.CASCADE, related_name='+')
    url = models.ForeignKey(LanguagePhrase, blank=False,
                            null=False, on_delete=models.CASCADE, related_name='+')
    color = models.ForeignKey(
        Color, on_delete=models.CASCADE, null=True, blank=True, related_name='+')
    appears = models.DateTimeField(
        help_text=HELP_FILL_TZ)
    vanishes = models.DateTimeField(
        help_text=HELP_FILL_TZ)

    reminder_title = 'Registration'
    kind = 'link'

    class Meta:
        ordering = ('convention__ceremony_opening_time', 'ordering', '-pk', )

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


class AdditionalRules(Timestampable):
    convention = models.ForeignKey(ConventionEdition, blank=False, null=False,
                                   on_delete=models.CASCADE, related_name='additional_rules')
    name = models.ForeignKey(LanguagePhrase, blank=False,
                             null=False, on_delete=models.CASCADE, related_name='+')
    rules = models.ForeignKey(LanguageRichText, blank=False,
                              null=False, on_delete=models.CASCADE, related_name='+')
    hidden = models.BooleanField(default=False, blank=True, null=False)
    ordering = models.IntegerField(default=0, blank=True, null=False)

    def __str__(self):
        return f'Additional rules #{self.pk} @ {self.convention.name.en}: {self.name.en}'

    class Meta:
        ordering = ('ordering', '-pk',)

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()


class UserAuthorization(Timestampable):
    user = models.ForeignKey(User, blank=False, null=False,
                             on_delete=models.CASCADE, related_name='authorizations')
    convention_series = models.ForeignKey(
        ConventionSeries, blank=False, null=False, on_delete=models.CASCADE, related_name='authorizations')
    is_owner = models.BooleanField(default=False, blank=True, null=False)

    def _delete_json_cache(self):
        if self.convention_series is None:
            super()._delete_json_cache()
        else:
            self.convention_series._delete_json_cache()


class SyncFailure(Timestampable):
    _uncollectable = True
    _strength = True
    synced_thing = models.TextField(
        default='', null=False, blank=True, primary_key=False, unique=True)
    count = models.BigIntegerField(
        default=0, null=False, blank=True, primary_key=False, unique=False)

    class Meta:
        indexes = [
            models.Index(fields=('synced_thing',),
                         name='sync_failure_thing_idx'),
        ]

    def _delete_json_cache(self):
        return None

    @classmethod
    def delete_if_exists(cls, idtf: str) -> bool:
        if (matched := cls.objects.filter(synced_thing=idtf).first()) is not None:
            matched.delete()
            return True
        return False

    @classmethod
    def get_or_create(cls, idtf: str) -> 'SyncFailure':
        if (matched := cls.objects.filter(synced_thing=idtf).first()) is not None:
            return matched
        item = SyncFailure(synced_thing=idtf, count=0)
        item.save()
        return item

    @classmethod
    def get_or_create_and_increment(cls, idtf: str) -> 'SyncFailure':
        item = cls.get_or_create(idtf)
        item.count += 1
        item.save()
        return item


class SummaryLink(Timestampable):
    ordering = models.IntegerField(default=0)
    convention = models.ForeignKey(ConventionEdition, default=None, blank=False,
                                   null=False, on_delete=models.CASCADE, related_name='summary_links')
    name = models.ForeignKey(LanguagePhrase, default=None, blank=False,
                             null=False, on_delete=models.CASCADE, related_name='+')
    description = models.ForeignKey(LanguagePhrase, default=None, blank=True,
                                    null=True, on_delete=models.CASCADE, related_name='+')
    icon = models.CharField(default='ic_ticket', blank=False,
                            null=False, max_length=255, choices=ALL_ICONS)
    url = models.ForeignKey(LanguagePhrase, default=None, blank=False,
                            null=False, on_delete=models.CASCADE, related_name='+')
    hidden = models.BooleanField(default=False, blank=True, null=False)

    def _delete_json_cache(self):
        if self.convention is None:
            super()._delete_json_cache()
        else:
            self.convention._delete_json_cache()

    class Meta:
        ordering = ('ordering', '-pk',)
