import datetime
import html
import sys
from hashlib import sha1
from textwrap import indent
from typing import Callable, Dict, List, Optional, Tuple, TypeAlias
from urllib import request

import bs4
import html2text as html2textlib
import ics
import requests
from slugify import slugify as slugify_

from . import models
from .hashcolor import get_rgbhex as obj2color
from .hashcolor import get_rgbhex_remindable_text as strml2color
from .helpers import (get_color_remidable_from_reminder_updating,
                      get_string_remidable_from_reminder_updating)

optstr = Optional[str]

PRESENTER_INTRODUCER_LINES = 'com ,con ,with '.split(',')
TRUE = True
HTTP_TIMEOUT = 120


def slugify(text: str, ok: str) -> str:
    return slugify_(text, ok=ok)


def ymdhm_different(ta: datetime.datetime, tb: datetime.datetime) -> bool:
    return any(map(lambda c: getattr(ta, c) != getattr(tb, c),
                   'year,month,day,hour,minute'.split(',')))


def join_lines_with_spaces_stripping(lines: List[str]):
    return ' '.join(map(str.strip, lines))


def split_title_with_presenter_introducer_line(multiline_text: str) -> Tuple[str, Optional[str]]:
    text_lines = multiline_text.splitlines()
    segregation_line: Optional[int] = None
    for line_idx, line in enumerate(text_lines):
        if line_idx == 0:
            continue
        if any(map(line.strip().lower().startswith, PRESENTER_INTRODUCER_LINES)):
            segregation_line = line_idx
            break
    if segregation_line is None:
        return join_lines_with_spaces_stripping(text_lines), None
    return (
        join_lines_with_spaces_stripping(text_lines[:segregation_line]),
        join_lines_with_spaces_stripping(text_lines[segregation_line:]),
    )


def html2text(html, with_md=True):
    h = html2textlib.HTML2Text(baseurl="", bodywidth=65535)
    if not with_md:
        h.ignore_emphasis = True
        h.ignore_images = True
        h.ignore_links = True
        h.ignore_tables = True
    return h.handle(html)


def textual_joiner(lst, mid, end):
    joined = list()
    for k, v in enumerate(lst):
        joined.append(v)
        if k < len(lst)-2:
            joined.append(mid)
        elif k == len(lst)-2:
            joined.append(end)
    return ''.join(joined)


fake_headers = {
    'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36',
    # 'From': 'youremail@domain.com'  # This is another valid field
}


def remote_ics_sync(url: str, current: models.ConventionEdition, apply_tz=True, ignore_description=False, ignore_subtitle=False):
    calendar = ics.Calendar(requests.get(
        url, headers=fake_headers, timeout=(HTTP_TIMEOUT, HTTP_TIMEOUT)).text)
    events: List[ics.Event] = list(calendar.events)
    known_activities: List[models.ConventionActivity] = list(
        current.events.all())
    known_activities_uuids = set(
        [x.remote_uuid for x in known_activities if x.remote_uuid is not None])
    for event_ in events:
        event: ics.Event = event_
        del event_
        activity_needs_saving = False
        activity: models.ConventionActivity = None
        if event.uid in known_activities_uuids:
            activity_ = current.events.filter(remote_uuid=event.uid)
            activity = activity_.first()
            if activity_.count() > 1:
                activity_.exclude(pk=activity.pk).delete()
            del activity_
        else:
            activity_needs_saving = True
            activity = models.ConventionActivity()
            activity.convention = current
            activity.remote_uuid = event.uid
            activity.conbook_key = "0"
            activity.conbook_pages = "0"
            activity.language = current.language
        begintime = event.begin.datetime
        timezone = datetime.timedelta(
            hours=abs(current.timezone.hour), minutes=current.timezone.minute)
        if apply_tz:
            if current.timezone.hour >= 0:
                begintime += timezone
            else:
                begintime -= timezone
        if not hasattr(activity, 'time_start') or activity.time_start != begintime:
            activity.time_start = begintime
            activity_needs_saving = True
        endtime = event.end.datetime
        if apply_tz:
            if current.timezone.hour >= 0:
                endtime += timezone
            else:
                endtime -= timezone
        if not hasattr(activity, 'time_end') or activity.time_end != endtime:
            activity.time_end = endtime
            activity_needs_saving = True
        if activity_needs_saving:
            try:
                activity.save()
                activity_needs_saving = False
            except Exception:
                pass
        name_reminder = f"at_nm_{current.edition_of.url}_auto_{event.uid}"
        subtitle_reminder = f"at_st_{current.edition_of.url}_auto_{event.uid}"
        description_reminder = f"at_dc_{current.edition_of.url}_auto_{event.uid}"
        event_name_raw = event.name.strip()
        event_name = ''
        event_subname = ''
        # print(event_name_raw, event_name_raw.splitlines(),
        #       '--------------------------------')
        if not ignore_subtitle and len(enl := event_name_raw.splitlines()) > 1 and (any(map(enl[1].lower().startswith, map(str.lower, PRESENTER_INTRODUCER_LINES)))):
            event_name = enl[0].strip()
            event_subname = ''.join(enl[1:]).strip()
        else:
            event_name = event_name_raw.strip().replace('\n', ' ').replace('  ', ' ')
            event_subname = ''
        name = get_string_remidable_from_reminder_updating(
            models.LanguagePhrase, name_reminder, *([event_name]*3))
        subtitle = None
        description = None
        if not ignore_subtitle and event_subname != '' and (
            (getattr(activity, 'subtitle', None) is not None)
            or
            (getattr(activity, 'subtitle', None) is None and subtitle is None)
        ):
            subtitle = get_string_remidable_from_reminder_updating(
                models.LanguagePhrase, subtitle_reminder, *([event_subname] * 3))
        if not ignore_description and (getattr(activity, 'description', None) is not None):
            description = get_string_remidable_from_reminder_updating(
                models.LanguageText, description_reminder, *([event.description or '-'] * 3))
        elif description is None and getattr(activity, 'description', None) is None:
            # cannot ignore_description and leave it blank
            description = get_string_remidable_from_reminder_updating(
                models.LanguageText, description_reminder, *(['-'] * 3))
        if not hasattr(activity, 'title') or activity.title != name:
            activity.title = name
            activity_needs_saving = True
        if not ignore_subtitle and (not hasattr(activity, 'subtitle') or activity.subtitle != subtitle):
            activity.subtitle = subtitle
            activity_needs_saving = True
        if description is not None:
            if not hasattr(activity, 'description') or activity.description != description:
                activity.description = description
                activity_needs_saving = True
        if activity_needs_saving:
            activity.save()
            activity_needs_saving = False
        categories_ids = set([x.pk for x in activity.categories.all()])
        for category_ in event.categories:
            category = category_.strip()
            sc = slugify(category, ok='_-')
            cat_reminder = f"cat_{current.edition_of.url}_auto_{sc}"
            clr_reminder = f"clr_{current.edition_of.url}_auto_{sc}"
            cat: models.ActivityCategory = None
            if models.ActivityCategory.objects.filter(name__en=category).exists():
                cat_ = models.ActivityCategory.objects.filter(
                    name__en=category)
                cat = cat_.first()
                cat_.exclude(pk=cat.pk).delete()
                del cat_
            else:
                nm = get_string_remidable_from_reminder_updating(
                    models.LanguagePhrase, cat_reminder, *([category]*3))
                clr = get_color_remidable_from_reminder_updating(
                    models.Color, clr_reminder, obj2color(category))
                cat = models.ActivityCategory(name=nm, color=clr)
                cat.save()
            if cat.edition is None:
                cat.edition = current
                cat.save()
            if cat.pk in categories_ids:
                categories_ids.remove(cat.pk)
            else:
                activity.categories.add(cat)
                activity_needs_saving = True
        for ctid in categories_ids:
            activity.categories.remove(ctid)
            activity_needs_saving = True
        if activity_needs_saving:
            activity.save()
            activity_needs_saving = False
        if TRUE:
            location = event.location.split(',')[0].strip()
            lc = slugify(location, ok='_-')
            loc_reminder = f"plc_{current.edition_of.url}_auto_{lc}"
            clr_reminder = f"clr_{current.edition_of.url}_auto_{lc}"
            loc: models.Place = models.Place.objects.filter(
                name__en=location).first()
            if loc is not None:
                models.Place.objects.filter(name__en=location).exclude(
                    pk=loc.pk).delete()
            else:
                nm = get_string_remidable_from_reminder_updating(
                    models.LanguagePhrase, loc_reminder, *([location]*3))
                clr = get_color_remidable_from_reminder_updating(
                    models.Color, clr_reminder, obj2color(location))
                loc = models.Place(name=nm, color=clr)
                loc.save()
            if loc.edition is None:
                loc.edition = current
                loc.save()
            places = list(activity.places.all())
            if len(places) != 1:
                activity.places.set([loc])
                activity_needs_saving = True
            else:
                place = places[0]
                if place.name.en != location:
                    activity.places.set([loc])
                    activity_needs_saving = True
        if activity_needs_saving:
            activity.save()
            activity_needs_saving = False
        if TRUE:
            desired_tiers = set(
                [x.pk for x in current.registration_tiers.all() if not x.is_lowest])
            attendable_saved = set(
                [x.pk for x in activity.attendable_by.all()])
            if desired_tiers != attendable_saved:
                activity.attendable_by.set(list(desired_tiers))
                activity_needs_saving = True
        if activity_needs_saving:
            activity.save()
            activity_needs_saving = False
        if event.classification is not None:
            if (clf := event.classification.lower()) in list(dict(models.TRILINGUAL_CHOICES).keys()):
                if activity.language != clf:
                    activity.language = clf
                    activity_needs_saving = True
        if activity_needs_saving:
            activity.save()
            activity_needs_saving = False
        try:
            known_activities_uuids.remove(event.uid)
        except Exception:
            pass
    cancelled_activities = current.events.filter(
        remote_uuid__in=known_activities_uuids)
    print(
        f"'{current.name.en}' has cancelled {cancelled_activities.count()} activities")
    cancelled_activities.delete()


def remote_sched_sync(name: str, current: models.ConventionEdition):
    remote_ics_sync(f"https://{name}.sched.com/all.ics",
                    current, ignore_description=True, ignore_subtitle=True)
    sched_html_raw = requests.get(
        f"https://{name}.sched.com/list/descriptions/", headers=fake_headers, timeout=(HTTP_TIMEOUT, HTTP_TIMEOUT)).text
    sched_html_bs = bs4.BeautifulSoup(sched_html_raw, "html5lib")
    sched_events_bs: List[bs4.Tag] = sched_html_bs.select(
        "#sched-content-inner div.sched-container div.sched-container-inner")
    for sched_event_bs_ in sched_events_bs:
        sched_event_bs: bs4.Tag = sched_event_bs_
        del sched_event_bs_
        uid = sched_event_bs.select_one('.event a.name')['id']
        description = None
        description_bs = sched_event_bs.select_one('.tip-description')
        if description_bs is not None:
            description = html2text(description_bs.prettify()).strip()
        people: List[str] = list()
        people_bs: bs4.Tag = sched_event_bs.select_one(
            '.sched-event-details-roles')
        if people_bs is not None:
            peoplelist_bs: List[bs4.Tag] = sched_event_bs.select('a[title]')
            peoplelist: List[str] = list(
                map(lambda a: a['title'].strip(), peoplelist_bs))
            people += peoplelist
        people = [person for person in people if
                  person != current.edition_of.name.en and
                  person != current.edition_of.name.es and
                  person != current.edition_of.name.pt
                  ]
        people_pt = "com "+textual_joiner(people, ', ', ' e ')
        people_en = "with "+textual_joiner(people, ', ', ' & ')
        people_es = "con "+textual_joiner(people, ', ', ' y ')
        activity = models.ConventionActivity.objects.get(remote_uuid=uid)
        actdescription = activity.description
        description_needs_saving = False
        if description is not None:
            if not hasattr(actdescription, 'en') or actdescription.en != description:
                actdescription.en = description
                description_needs_saving = True
            if not hasattr(actdescription, 'es') or actdescription.es != description:
                actdescription.es = description
                description_needs_saving = True
            if not hasattr(actdescription, 'pt') or actdescription.pt != description:
                actdescription.pt = description
                description_needs_saving = True
            if description_needs_saving:
                actdescription.save()
                description_needs_saving = False
        else:
            print(
                f"'{current.name.en}' has '{activity.title.en}' activity with empty description")
        subtitle_reminder = f"at_st_{current.edition_of.url}_auto_{uid}"
        subtitle = None
        subtitle_needs_saving = False
        if len(people) == 0:
            models.LanguagePhrase.objects.filter(
                reminder_alias=subtitle_reminder).delete()
            if hasattr(activity, 'subtitle') and activity.subtitle is not None:
                activity.subtitle = None
                activity.save()
        else:
            if hasattr(activity, 'subtitle') and activity.subtitle is not None:
                subtitle = activity.subtitle
            elif models.LanguagePhrase.objects.filter(reminder_alias=subtitle_reminder).exists():
                subtitle_needs_saving = True
                subtitle = models.LanguagePhrase.objects.get(
                    reminder_alias=subtitle_reminder)
            else:
                subtitle_needs_saving = True
                subtitle = models.LanguagePhrase(
                    reminder_alias=subtitle_reminder,
                    pt=people_pt,
                    en=people_en,
                    es=people_es
                )
            if not hasattr(subtitle, 'en') or subtitle.en != people_en:
                subtitle.en = people_en
                subtitle_needs_saving = True
            if not hasattr(subtitle, 'es') or subtitle.es != people_es:
                subtitle.es = people_es
                subtitle_needs_saving = True
            if not hasattr(subtitle, 'pt') or subtitle.pt != people_pt:
                subtitle.pt = people_pt
                subtitle_needs_saving = True
            if subtitle_needs_saving:
                activity.subtitle = subtitle
                subtitle.save()
                activity.save()


def remote_announcement_telegram_sync(convention: models.ConventionSeries):
    tch = convention.telegram_channel
    if tch is None or tch == '':
        return
    channel_raw = requests.get(
        f'https://t.me/s/{tch}', headers=fake_headers, timeout=(HTTP_TIMEOUT, HTTP_TIMEOUT)).text
    channel_bs = bs4.BeautifulSoup(channel_raw, 'html5lib')
    messages_bs: List[bs4.Tag] = channel_bs.select('.tgme_widget_message_wrap')
    for message_bs in messages_bs:
        body_bs = message_bs.select_one('.tgme_widget_message_text')
        if body_bs is None:
            continue
        message_id = int(message_bs.select_one(
            'a.tgme_widget_message_date[href]')['href'].split('/')[-1])
        for tag_ in body_bs.select('a[href]'):
            tag: bs4.Tag = tag_
            del tag_
            if tag.text.strip() == tag['href'].strip():
                tag.replaceWith(tag['href'].strip())
        body = html2text(body_bs.prettify(), False).strip()
        title_en = "Announcement forwarded from Telegram"
        title_es = "Anuncio reenviado desde Telegram"
        title_pt = "Comunicado encaminhado do Telegram"
        lines = body.splitlines()
        if len(lines) > 1:
            title_en = lines[0]
            title_es = lines[0]
            title_pt = lines[0]
            new_body = '\n'.join(lines[1:]).strip()
            body = body if len(new_body.splitlines()) <= 0 else new_body
        link = f'https://t.me/s/{tch}/{message_id}'
        announcement_reminder = f'notif_{convention.url}_auto_telegram_channel_{message_id}'
        title_reminder = f'{announcement_reminder}_title'
        body_reminder = f'{announcement_reminder}_body'
        link_reminder = f'{announcement_reminder}_link'
        if not models.Announcement.objects.filter(telegram_id=message_id).exists():
            title_obj = get_string_remidable_from_reminder_updating(
                models.LanguagePhrase, title_reminder, title_en, title_es, title_pt)
            body_obj = get_string_remidable_from_reminder_updating(
                models.LanguageText, body_reminder, *([body]*3))
            link_obj = get_string_remidable_from_reminder_updating(
                models.LanguagePhrase, link_reminder, *([link]*3))
            announcement = models.Announcement(
                convention=convention,
                title=title_obj,
                body=body_obj,
                link=link_obj,
                telegram_id=message_id
            )
            announcement.save()


def remote_htmltable2json_sync(url: str, current: models.ConventionEdition):
    resp = requests.get(url, timeout=(HTTP_TIMEOUT, HTTP_TIMEOUT))
    resp.raise_for_status()
    activities_json = resp.json()
    del resp
    activities_ids = [x['uid'] for x in activities_json]
    existing_events = current.events.filter(
        remote_uuid__in=activities_ids)
    excluded_events = current.events.all().exclude(
        remote_uuid__in=activities_ids)
    print(f"'{current.name.en}' has cancelled {excluded_events.count()} activities")
    excluded_events.delete()
    print(f"'{current.name.en}' will create {len(activities_ids)-existing_events.count()} new activities and update {existing_events.count()} if needed")
    del activities_ids
    del excluded_events
    del existing_events
    stwpil = split_title_with_presenter_introducer_line
    for activity_json in activities_json:
        activity_uuid: str = activity_json['uid']
        activity_start: datetime.datetime = datetime.datetime.fromisoformat(
            activity_json['start'])
        activity_end: datetime.datetime = datetime.datetime.fromisoformat(
            activity_json['end'])
        activity_place_en: str = activity_json['place']['en']
        activity_place_es: str = activity_json['place']['es']
        activity_place_pt: str = activity_json['place']['pt']
        activity_place_id: int = activity_json['place']['id']
        activity_name_en: str = stwpil(activity_json['name']['en'])[0]
        activity_name_es: str = stwpil(activity_json['name']['es'])[0]
        activity_name_pt: str = stwpil(activity_json['name']['pt'])[0]
        activity_subtitle_en: optstr = stwpil(activity_json['name']['en'])[1]
        activity_subtitle_es: optstr = stwpil(activity_json['name']['es'])[1]
        activity_subtitle_pt: optstr = stwpil(activity_json['name']['pt'])[1]
        activity_description_en: str = activity_json['description']['en']
        activity_description_es: str = activity_json['description']['es']
        activity_description_pt: str = activity_json['description']['pt']
        activity_language: str = activity_json['language']
        name_reminder = f"at_nm_{current.edition_of.url}_auto_{activity_uuid}"
        subtitle_reminder = f"at_st_{current.edition_of.url}_auto_{activity_uuid}"
        description_reminder = f"at_dc_{current.edition_of.url}_auto_{activity_uuid}"
        subtitle_ml: Optional[models.LanguagePhrase] = None
        if not all([activity_subtitle_en, activity_subtitle_es, activity_subtitle_pt]):
            activity_name_en = join_lines_with_spaces_stripping([
                activity_name_en,
                activity_subtitle_en or '',
            ])
            activity_subtitle_en = None
            activity_name_es = join_lines_with_spaces_stripping([
                activity_name_es,
                activity_subtitle_es or '',
            ])
            activity_subtitle_es = None
            activity_name_pt = join_lines_with_spaces_stripping([
                activity_name_pt,
                activity_subtitle_pt or '',
            ])
            activity_subtitle_en = None
        else:
            subtitle_ml = get_string_remidable_from_reminder_updating(
                models.LanguagePhrase,
                subtitle_reminder,
                activity_subtitle_en or '',
                activity_subtitle_es or '',
                activity_subtitle_pt or '',
            )
        name_ml = get_string_remidable_from_reminder_updating(
            models.LanguagePhrase,
            name_reminder,
            activity_name_en,
            activity_name_es,
            activity_name_pt,
        )
        description_ml = get_string_remidable_from_reminder_updating(
            models.LanguageText,
            description_reminder,
            activity_description_en,
            activity_description_es,
            activity_description_pt,
        )
        place_name_reminder = f"plc_{current.edition_of.url}_auto_{activity_place_id}"
        place_color_reminder = f"clr_{current.edition_of.url}_auto_{activity_place_id}"
        place_name_ml = get_string_remidable_from_reminder_updating(
            models.LanguagePhrase,
            place_name_reminder,
            activity_place_en,
            activity_place_es,
            activity_place_pt,
        )
        place_color_ml = get_color_remidable_from_reminder_updating(
            models.Color, place_color_reminder, strml2color(place_name_ml))
        place_ml: models.Place = models.Place.objects.filter(
            edition=current, name=place_name_ml, color=place_color_ml).first()
        if place_ml is None:
            place_ml = models.Place(
                edition=current, name=place_name_ml, color=place_color_ml)
            place_ml.save()
        attendables = [x for x in current.registration_tiers.all()
                       if not x.is_lowest]
        if (activity_ := current.events.filter(remote_uuid=activity_uuid).first()):
            activity: models.ConventionActivity = activity_
            del activity_
            activity_changed = False
            if getattr(activity.title, 'pk', -1) != getattr(name_ml, 'pk', -1):
                activity.title = name_ml
                activity_changed = True
            if getattr(activity.subtitle, 'pk', -1) != getattr(subtitle_ml, 'pk', -1):
                activity.subtitle = subtitle_ml
                activity_changed = True
            if getattr(activity.description, 'pk', -1) != getattr(description_ml, 'pk', -1):
                activity.description = description_ml
                activity_changed = True
            if (activity.language or '') != activity_language.lower():
                activity.language = activity_language.lower()
                activity_changed = True
            if (
                ymdhm_different(activity.time_start, activity_start)
                or
                ymdhm_different(activity.time_end, activity_end)
            ):
                activity.time_start = datetime.datetime(activity_start.year,
                                                        activity_start.month,
                                                        activity_start.day,
                                                        activity_start.hour,
                                                        activity_start.minute,)
                activity.time_end = datetime.datetime(activity_end.year,
                                                      activity_end.month,
                                                      activity_end.day,
                                                      activity_end.hour,
                                                      activity_end.minute,)
                activity_changed = True
            if activity.categories.count():
                activity.categories.set([])
                activity_changed = True
            if set(x.pk for x in attendables) != set(x.pk for x in activity.attendable_by.all()):
                activity.attendable_by.set(attendables)
                activity_changed = True
            if set([place_ml.pk]) != set(x.pk for x in activity.places.all()):
                activity.places.set([place_ml])
                activity_changed = True
            if activity_changed:
                activity_changed = False
                activity.save()
        else:
            activity = models.ConventionActivity(
                is_draft=False,
                conbook_key='0',
                conbook_pages='0',
                picture=None,
                title=name_ml,
                subtitle=subtitle_ml,
                description=description_ml,
                time_start=datetime.datetime(activity_start.year,
                                             activity_start.month,
                                             activity_start.day,
                                             activity_start.hour,
                                             activity_start.minute,),
                time_end=datetime.datetime(activity_end.year,
                                           activity_end.month,
                                           activity_end.day,
                                           activity_end.hour,
                                           activity_end.minute,),
                language=activity_language.lower(),
                hidden_from_time_table=False,
                convention=current,
                remote_uuid=activity_uuid
            )
            activity.save()
            activity.attendable_by.set(attendables)
            activity.places.set([place_ml])
            activity.categories.set([])
            activity.save()


def http_get_returns_200(url: str) -> bool:
    try:
        response = requests.get(url, timeout=(HTTP_TIMEOUT, HTTP_TIMEOUT))
        if '<h1>Please wait while your request is being verified...</h1>' in response.text:
            return False
        return response.status_code == 200
    except Exception:
        return False


schedule_sync_handlers: Dict[str, Callable[[str, models.ConventionEdition], None]] = {
    'ics': remote_ics_sync,
    'sched': remote_sched_sync,
    'htmltable2json': remote_htmltable2json_sync,
}

announcement_sync_handlers: Dict[str, Callable[[models.ConventionEdition], None]] = {
    'telegram_channel': remote_announcement_telegram_sync,
}


def update_language_from(lang: models.SingleChildLanguagePhraseCopyable, url: str):
    resp = requests.get(url, timeout=(HTTP_TIMEOUT, HTTP_TIMEOUT))
    resp.raise_for_status()
    dt = resp.json()
    lang_en = dt['en']
    lang_es = dt['es']
    lang_pt = dt['pt']
    updated = False
    if isinstance(lang_en, str) and lang_en != lang.en:
        updated = True
        lang.en = lang_en
    if isinstance(lang_es, str) and lang_es != lang.es:
        updated = True
        lang.es = lang_es
    if isinstance(lang_pt, str) and lang_pt != lang.pt:
        updated = True
        lang.pt = lang_pt
    if updated:
        lang.save()
